/*
	Code (C) 2005 Alexander A. Seleznyov aseleznyov@gmail.com http://www.seleznyov.com
	Use freely. Try to keep copyright info untouched if possible :)
*/
#pragma once

#ifndef __ATLSMTPAUTHCONNECTION_H__
#define __ATLSMTPAUTHCONNECTION_H__

#ifndef __ATLSMTPCONNECTION_H__
#include <atlsmtpconnection.h>
#endif

#ifndef __ATLENC_H__
#include <atlenc.h>
#endif

//#define USENTML

#ifdef USENTLM
#ifndef SECURITY_WIN32
#define SECURITY_WIN32
#define WIN32_CHICAGO
#include <security.h>
#endif

#pragma comment(lib, "Secur32.lib")
#endif

#define ATLSMTP_EHLO_SUCC "250"
#define ATLSMTP_ALOG_SUCC "334"
#define ATLSMTP_ALOG_DONE "235"
#define ATLSMTP_NTLM_SUCC "334"
#define ATLSMTP_NTLM_DONE "235"

namespace ATL
{
	class CSMTPAuthConnection : public CSMTPConnection
	{
	public:
		CSMTPAuthConnection() throw(): CSMTPConnection() 
		{
			ATLTRACE("CSMTPAuthConnection::CSMTPAuthConnection() called\n");
		}
		// Attempt to connect to the socket
		// lpszHostName - the host name to connect to
		BOOL ConnectESMTP(LPCTSTR lpszHostName, DWORD dwTimeout = 10000) throw()
		{
			ATLASSERT(lpszHostName != NULL);

			// If we're already connected
			if (Connected())
			{
				return FALSE;
			}

			if (!_g_smtp_init.Init())
			{
				return FALSE;
			}

			CSocketAddr address;
			if (address.FindAddr(lpszHostName, IPPORT_SMTP, 0, PF_UNSPEC, SOCK_STREAM, 0))
			{
				return FALSE;
			}

			ADDRINFOT *pAI;

			BOOL bRet = FALSE;
			int nIndex = 0;
			while ((pAI = address.GetAddrInfo(nIndex++)) != NULL)
			{
				// create the socket
				m_hSocket = WSASocket(pAI->ai_family, pAI->ai_socktype, pAI->ai_protocol, NULL, 0, WSA_FLAG_OVERLAPPED);

				if (m_hSocket == INVALID_SOCKET)
				{
					return FALSE;
				}

				bRet = FALSE;
				WSAEVENT hEventConnect = WSACreateEvent();
				if (hEventConnect != NULL)
				{
					if (SOCKET_ERROR != WSAEventSelect(m_hSocket, hEventConnect, FD_CONNECT))
					{
						if (WSAConnect(m_hSocket, pAI->ai_addr, (int)pAI->ai_addrlen, 
							NULL, NULL, NULL, NULL))
						{
							if (WSAGetLastError() == WSAEWOULDBLOCK)
							{
								DWORD dwWait = WaitForSingleObject((HANDLE) hEventConnect, dwTimeout);
								if (dwWait == WAIT_OBJECT_0)
								{
									// make sure there were no connection errors.
									WSANETWORKEVENTS wse;
									ZeroMemory(&wse, sizeof(wse));
									WSAEnumNetworkEvents(m_hSocket, NULL, &wse);
									if (wse.iErrorCode[FD_CONNECT_BIT]==0)
									{
										bRet = TRUE;
									}
								}
							}
						}
						else
							bRet = TRUE;
					}

					// we're done with the event
					WSACloseEvent(hEventConnect);
				}
				if (bRet)
				{
					break;
				}

				shutdown(m_hSocket, SD_BOTH);
				closesocket(m_hSocket);
				m_hSocket = INVALID_SOCKET;
			}


			// Create an event for asynchronous I/O
			if (bRet)
			{
				ATLASSERT(m_Overlapped.hEvent == NULL);
				m_Overlapped.hEvent = CreateEvent(NULL, TRUE, TRUE, NULL);
				if (m_Overlapped.hEvent == NULL)
				{
					bRet = FALSE;
				}
			}

			char szBuf[ATLSMTP_MAX_LINE_LENGTH+1];
			int nBufLen = ATLSMTP_MAX_LINE_LENGTH;
			if (bRet)
			{
				// See if the connect returns success
				bRet = AtlSmtpReadData((HANDLE)m_hSocket, szBuf, &nBufLen, &m_Overlapped);
				if (bRet)
				{
					if (strncmp(szBuf, ATLSMTP_CONN_SUCC, ATLSMTP_RETCODE_LEN))
					{
						bRet = FALSE;
					}
				}
			}

			char szLocalHost[ATLSMTP_MAX_SERVER_NAME_LENGTH+1];

			// gethostname should return 0 on success
			if (bRet && gethostname(szLocalHost, ATLSMTP_MAX_SERVER_NAME_LENGTH))
			{
				bRet = FALSE;
			}

			// Send EHLO command and get reply
			if (bRet)
			{
				nBufLen = _snprintf(szBuf, ATLSMTP_MAX_LINE_LENGTH+1, "EHLO %s\r\n", szLocalHost);
				if (nBufLen > 0)
				{
					bRet = AtlSmtpSendAndCheck((HANDLE)m_hSocket, szBuf, nBufLen, szBuf, &nBufLen, 
						ATLSMTP_MAX_LINE_LENGTH, ATLSMTP_HELO_SUCC, &m_Overlapped);
				}
				else
				{
					bRet = FALSE;
				}
			}

			if (!bRet)
			{
				if (m_Overlapped.hEvent != NULL)
					CloseHandle(m_Overlapped.hEvent);
				shutdown(m_hSocket, SD_BOTH);
				closesocket(m_hSocket);
				m_hSocket = INVALID_SOCKET;
			}

			return bRet;
		}
		// Authenticate by AUTH LOGIN
		BOOL AuthenticateLogin(LPCTSTR lpszLogin, LPCTSTR lpszPassword) throw()
		{
			ATLASSERT(lpszLogin);
			
			char szBuf[ATLSMTP_MAX_LINE_LENGTH+1];
			BOOL bRet;

			int nBufLen = _snprintf(szBuf, ATLSMTP_MAX_LINE_LENGTH+1, "AUTH LOGIN\r\n");
			if (nBufLen > 0)
			{
				bRet = AtlSmtpSendAndCheck((HANDLE)m_hSocket, szBuf, nBufLen, szBuf, &nBufLen, 
					ATLSMTP_MAX_LINE_LENGTH, ATLSMTP_ALOG_SUCC, &m_Overlapped);
				if (bRet)
					szBuf[nBufLen] = 0; // append terminator character
			}
			else
			{
				bRet = FALSE;
			}
			if (bRet) // proceed with username
			{
				char szBufB64[ATLSMTP_MAX_LINE_LENGTH+1];
				int nBufB64;
				nBufB64 = ATLSMTP_MAX_LINE_LENGTH+1;
				bRet = Base64Decode(szBuf+4, nBufLen+4, (LPBYTE)szBufB64, &nBufB64);
				bRet = strnicmp("username:", szBufB64, 9)==0;
				if (bRet) // send encoded user name
				{
					nBufB64 = ATLSMTP_MAX_LINE_LENGTH+1;
					Base64Encode((LPBYTE)lpszLogin, (int)strlen(lpszLogin), szBufB64, &nBufB64, ATL_BASE64_FLAG_NOCRLF);
					szBufB64[nBufB64] = 0;
					_snprintf(szBuf, ATLSMTP_MAX_LINE_LENGTH+1, "%s\r\n", szBufB64);
					nBufLen = (int)strlen(szBuf);
					bRet = AtlSmtpSendAndCheck((HANDLE)m_hSocket, szBuf, nBufLen, szBuf, &nBufLen, 
						ATLSMTP_MAX_LINE_LENGTH, ATLSMTP_ALOG_SUCC, &m_Overlapped);
				}
			}
			if (bRet) // proceed with password
			{
				char szBufB64[ATLSMTP_MAX_LINE_LENGTH+1];
				int nBufB64;
				nBufB64 = ATLSMTP_MAX_LINE_LENGTH+1;
				bRet = Base64Decode(szBuf+4, nBufLen+4, (LPBYTE)szBufB64, &nBufB64);
				bRet = strnicmp("password:", szBufB64, 9)==0;
				if (bRet) // send encoded password
				{
					nBufB64 = ATLSMTP_MAX_LINE_LENGTH+1;
					Base64Encode((LPBYTE)lpszPassword, (int)strlen(lpszPassword), szBufB64, &nBufB64, ATL_BASE64_FLAG_NOCRLF);
					szBufB64[nBufB64] = 0;
					_snprintf(szBuf, ATLSMTP_MAX_LINE_LENGTH+1, "%s\r\n\0", szBufB64);
					nBufLen = (int)strlen(szBuf);
					bRet = AtlSmtpSendAndCheck((HANDLE)m_hSocket, szBuf, nBufLen, szBuf, &nBufLen, 
						ATLSMTP_MAX_LINE_LENGTH, ATLSMTP_ALOG_DONE, &m_Overlapped);
				}
			}
			
			return bRet;
		}
#ifdef USENTLM
		BOOL AuthenticateNTLM() throw()
		{
			char szBuf[16384]; // will be used for response
			int nBuf;
			char szBuf2[16384];
			int nBuf2;

			BOOL done = FALSE;

			CredHandle hCred;
			SecHandle hSec;

			ZeroMemory(&hCred, sizeof(CredHandle));
			ZeroMemory(&hSec, sizeof(SecHandle));

			BYTE abIn[12000];
			BYTE abOut[12000];

			DWORD cbIn = sizeof(abIn);
			DWORD cbOut = sizeof(abOut);

			SECURITY_STATUS status = GenClientContext(NULL, 0, abOut, &cbOut, &done, &hCred, &hSec);
			if (SEC_SUCCESS(status))
			{
				strcpy(szBuf2, "AUTH NTLM ");
				nBuf2 = sizeof(szBuf2)-10;
				Base64Encode(abOut, cbOut, szBuf2+10, &nBuf2, ATL_BASE64_FLAG_NOCRLF);
				nBuf2 += 10;
				strcpy(&(szBuf2[nBuf2]), "\r\n");
				nBuf2 += 2;
				nBuf = sizeof(szBuf);
				status = AtlSmtpSendAndCheck((HANDLE)m_hSocket, szBuf2, nBuf2, szBuf, &nBuf, 
					ATLSMTP_MAX_LINE_LENGTH, ATLSMTP_NTLM_SUCC, &m_Overlapped) ? SEC_E_OK : ERROR_BAD_NET_RESP;
			}
			while (!done && SEC_SUCCESS(status))
			{
				if (SEC_SUCCESS(status))
				{
					status = Base64Decode(szBuf+4, nBuf-4, abIn, (int *)&cbIn) ? SEC_E_OK : ERROR_BAD_NET_RESP;
				}
				if (SEC_SUCCESS(status))
				{
					cbOut = sizeof(abOut);
					status = GenClientContext(abIn, cbIn, abOut, &cbOut, &done, &hCred, &hSec);
				}
				else
					break;
				if (SEC_SUCCESS(status))
				{
					nBuf2 = sizeof(szBuf2);
					Base64Encode(abOut, cbOut, szBuf2, &nBuf2, ATL_BASE64_FLAG_NOCRLF);
					strcpy(&(szBuf2[nBuf2]), "\r\n");
					nBuf2 += 2;
					status = AtlSmtpSendAndCheck((HANDLE)m_hSocket, szBuf2, nBuf2, szBuf, &nBuf, 
						ATLSMTP_MAX_LINE_LENGTH, ATLSMTP_NTLM_DONE, &m_Overlapped) ? SEC_E_OK : ERROR_BAD_NET_RESP;
				}
				else
					break;
			}
			//Free up the security context if valid
			if (hSec.dwLower || hSec.dwUpper) 
				DeleteSecurityContext(&hSec);

			//Free up the credentials handle if valid
			if (hCred.dwLower || hCred.dwUpper) 
				FreeCredentialHandle(&hCred);

			return status==SEC_E_OK;
		}
	protected:
		SECURITY_STATUS GenClientContext(LPBYTE pIn, DWORD cbIn, LPBYTE pOut, LPDWORD pcbOut, LPBOOL pfDone, CredHandle* phCred, SecHandle* phContext) throw()
		{
			TimeStamp Lifetime;
			SECURITY_STATUS ss;
			if (NULL == pIn)  
			{   
				ss = AcquireCredentialsHandle(NULL, _T("NTLM"), SECPKG_CRED_OUTBOUND, NULL, NULL, NULL, NULL, phCred, &Lifetime);
				if (!SEC_SUCCESS(ss))
					return ss;
			}

			//Prepare the buffers
			SecBufferDesc OutBuffDesc;
			OutBuffDesc.ulVersion = 0;
			OutBuffDesc.cBuffers  = 1;
			SecBuffer OutSecBuff;
			OutBuffDesc.pBuffers  = &OutSecBuff;
			OutSecBuff.cbBuffer   = *pcbOut;
			OutSecBuff.BufferType = SECBUFFER_TOKEN;
			OutSecBuff.pvBuffer   = pOut;

			//The input buffer is created only if a message has been received from the server.
			SecBufferDesc InBuffDesc;
			SecBuffer InSecBuff;
			if (pIn)   
			{
				InBuffDesc.ulVersion = 0;
				InBuffDesc.cBuffers  = 1;
				InBuffDesc.pBuffers  = &InSecBuff;
				InSecBuff.cbBuffer   = cbIn;
				InSecBuff.BufferType = SECBUFFER_TOKEN;
				InSecBuff.pvBuffer   = pIn;

				ULONG ContextAttributes;
				ss = InitializeSecurityContext(phCred, phContext, NULL, 0, 0, SECURITY_NATIVE_DREP, &InBuffDesc, 0, phContext, &OutBuffDesc, &ContextAttributes, &Lifetime);
			}
			else
			{
				ULONG ContextAttributes;
				ss = InitializeSecurityContext(phCred, NULL, NULL, 0, 0, SECURITY_NATIVE_DREP, NULL, 0, phContext, &OutBuffDesc, &ContextAttributes, &Lifetime);
			}

			if (!SEC_SUCCESS(ss))
				return ss;

			//If necessary, complete the token.
			if ((SEC_I_COMPLETE_NEEDED == ss) || (SEC_I_COMPLETE_AND_CONTINUE == ss))  
			{
				ss = CompleteAuthToken(phContext, &OutBuffDesc);
				if (!SEC_SUCCESS(ss))  
					return ss;
			}

			*pcbOut = OutSecBuff.cbBuffer;
			*pfDone = !((SEC_I_CONTINUE_NEEDED == ss) || (SEC_I_COMPLETE_AND_CONTINUE == ss));

			return ss;
		}
#endif
	};
	// CMimeText - represents a html body part in MIME body
	class CMimeHTML : public CMimeText
	{
	public:
		CMimeHTML() throw()
			: CMimeText()
		{
		}
		virtual inline LPCSTR GetContentType() throw()
		{
			return "text/html";
		}
	protected:
		// Make the MIME header
		virtual inline BOOL MakeMimeHeader(CStringA& header, LPCSTR szBoundary) throw()
		{
			char szBegin[256];
			if (*szBoundary)
			{
				// this is not the only body part
				memcpy(szBegin, "\r\n\r\n--", 6);
				memcpy(szBegin+6, szBoundary, ATL_MIME_BOUNDARYLEN);
				*(szBegin+(ATL_MIME_BOUNDARYLEN+6)) = '\0';
			}
			else
			{
				// this is the only body part, so output the full MIME header
				memcpy(szBegin, "MIME-Version: 1.0", sizeof("MIME-Version: 1.0"));
			}

			_ATLTRY
			{
				header.Format("%s\r\nContent-Type: %s;\r\n\tcharset=\"%s\"\r\nContent-Transfer-Encoding: 8bit\r\n\r\n", 
					szBegin, GetContentType(), m_szCharset);
				return TRUE;
			}
			_ATLCATCHALL()
			{
				return FALSE;
			}
		}
	}; // class CMimeHTML

	class CMimeMessage2 : public CMimeMessage
	{
	public:
		CMimeMessage2(IMultiLanguage *pMultiLanguage = NULL) throw()
			: CMimeMessage(pMultiLanguage)
		{
		}
		void AddRecipients(LPCTSTR lpszRecepients) throw()
		{
			CFixedStringT<CString, MAX_PATH> strRecip;
			LPCTSTR szTmp = lpszRecepients;
			LPCTSTR szTmp2 = lpszRecepients;
			BOOL bRet = TRUE;
			while (*szTmp && bRet)
			{
				if (AtlSmtpIsRecipientDelimiter((char) *szTmp2))
				{
					_ATLTRY
					{
						strRecip.SetString(szTmp, (int)(szTmp2-szTmp));
						bRet = AddRecipient((LPCTSTR) strRecip);

						if (*szTmp2)
						{
							while (*szTmp2 && AtlSmtpIsRecipientDelimiter((char) *szTmp2))
							{
								szTmp2++;
							}
						}
						szTmp = szTmp2;
					}
					_ATLCATCHALL()
					{
						bRet = FALSE;
					}
				}
				else
				{
					szTmp2++;
				}
			}
		}
		// Add html to the message at position nPos in the body parts list
		// szHtml - the html code
		// nTextLen - the size of the text in bytes (optional - if not specified a _tcslen will be done)
		// nPos - the position in the message at which to insert the text (optional)
		// uiCodePage - the codepage (optional)
		inline BOOL AddHTML(LPCTSTR szHtml, int nTextLen = -1, int nPos = 1, UINT uiCodePage = 0) throw()
		{
			if (szHtml == NULL)
				return FALSE;

			if (nPos < 1)
			{
				nPos = 1;
			}

			CAutoPtr<CMimeBodyPart> spNewHtml;
			CMimeHTML *pNewHtml = NULL;
			ATLTRY(spNewHtml.Attach(pNewHtml = new CMimeHTML()));
			if (!spNewHtml)
				return FALSE;

			BOOL bRet = pNewHtml->Initialize(szHtml, nTextLen, m_spMultiLanguage, uiCodePage);
			if (bRet)
			{
				_ATLTRY
				{
					POSITION currPos = m_BodyParts.FindIndex(nPos-1);

					if (!currPos)
					{
						if (!m_BodyParts.AddTail(spNewHtml))
							bRet = FALSE;
					}
					else
					{
						if (!m_BodyParts.InsertBefore(currPos, spNewHtml))
							bRet = FALSE;
					}

				}
				_ATLCATCHALL()
				{
					bRet = FALSE;
				}
			}

			return bRet;
		}
	};
}

#endif
